﻿using System.Collections.Concurrent;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Xml.Serialization;
using ff14bot.Helpers;
using ProtoBuf;

namespace Zekken.Avoidance
{
    public class SpellShapeCompiler
    {
        private static XmlSerializer xmlSerializer = new XmlSerializer(typeof (List<XSpellShape>));

        public static void CompileShapes()
        {
            if (!Directory.Exists(Common.DEFAULT_SHAPES_PATH))
            {
                Logging.Write(Common.ErrorColor, "[Zekken] Cannot compile, the spell shape directory doesn't exist.");
                return;
            }

            var files = GetXmlFilesFrom(Common.DEFAULT_SHAPES_PATH)
                .ToList();

            foreach (var file in files) { CleanFile(file); }

            var shapes = files.SelectMany(LoadShapes).ToList();

            foreach (var shape in shapes)
            {
                shape.Flags = shape.Flags.Distinct().ToList();
            }

            foreach (var file in files) { File.Delete(file); }

            var purgedShapes = GetPurgedShapes(shapes).OrderBy(s => s.Id).ToList();
            Logging.Write(Common.LogColor, string.Format("[Zekken] Found {0} unique spell shapes and {1} repetitions.", 
                purgedShapes.Count, shapes.Count - purgedShapes.Count));

            Save(Common.DEFAULT_XDB_PATH, purgedShapes);

            var compiledData = new ConcurrentDictionary<uint, SpellShape>();
            foreach (var shape in purgedShapes) { compiledData[shape.Id] = shape.GetShape(); }
            Save(compiledData);

            Logging.Write(Common.LogColor, string.Format("[Zekken] Finished compiling {0} spell shapes.", purgedShapes.Count));
        }

        public ConcurrentDictionary<uint, SpellShape> Load()
        {
            return Load(Common.DEFAULT_DATA_PATH);
        }

        public ConcurrentDictionary<uint, SpellShape> Load(string path)
        {
            if (!File.Exists(path))
            {
                Logging.Write(Common.ErrorColor, "[Zekken] Spell data file does not exist.");
                return new ConcurrentDictionary<uint, SpellShape>();
            }

            ConcurrentDictionary<uint, SpellShape> loadedShapes = null;
            using (var stream = new FileStream(path, FileMode.Open, FileAccess.Read))
            {
                try { loadedShapes = Serializer.Deserialize<ConcurrentDictionary<uint, SpellShape>>(stream); }
                catch { Logging.Write(Common.ErrorColor, "[Zekken] Failed to load spell shape data."); }
            }

            if (loadedShapes == null) { return new ConcurrentDictionary<uint, SpellShape>(); }

            Logging.Write(Common.LogColor, "[Zekken] Loaded spell shape data.");
            return loadedShapes;
        }

        public static bool Save(ConcurrentDictionary<uint, SpellShape> data)
        {
            return Save(Common.DEFAULT_DATA_PATH, data);
        }

        public static bool Save(string path, ConcurrentDictionary<uint, SpellShape> data)
        {
            var fileInfo = new FileInfo(path);

            if (fileInfo.Directory == null || !Directory.Exists(fileInfo.Directory.FullName))
            {
                Logging.Write(Common.ErrorColor, "[Zekken] Cannot save spell data, the provided directory does not exist.");
                return false;
            }

            if (fileInfo.Exists)
            {
                try { File.Delete(path); }
                catch
                {
                    Logging.Write(Common.ErrorColor, "[Zekken] Failed to overwrite spell shape data file.");
                    return false;
                }
            }

            using (var stream = new FileStream(path, FileMode.Create, FileAccess.Write))
            {
                try { Serializer.Serialize(stream, data); }
                catch
                {
                    Logging.Write(Common.ErrorColor, "[Zekken] Failed to save spell shape data.");
                    return false;
                }
            }

            Logging.Write(Common.LogColor, "[Zekken] Saved spell shape data to " + path);
            return true;
        }

        public static List<XSpellShape> LoadShapes(string path)
        {
            List<XSpellShape> shapes;

            using (var stream = new FileStream(path, FileMode.Open, FileAccess.Read)) 
            {
                try { shapes = (List<XSpellShape>) xmlSerializer.Deserialize(stream); }
                catch
                {
                    Logging.Write(Common.ErrorColor, string.Format("[Zekken] Could not deserialize shapes from {0}", path));
                    shapes = new List<XSpellShape>();
                }
            }

            return shapes;
        }

        public static void Save(string path, List<XSpellShape> shapes)
        {
            if (File.Exists(path))
            {
                try { File.Delete(path); }
                catch
                {
                    Logging.Write(Common.ErrorColor, "[Zekken] Failed to overwrite spell shape xml file.");
                    return;
                }
            }

            using (var stream = new FileStream(path, FileMode.Create, FileAccess.Write)) 
            {
                xmlSerializer.Serialize(stream, shapes);
            }
        }

        public static IEnumerable<string> GetXmlFilesFrom(string directory)
        {
            return Directory.EnumerateFiles(directory, "*.xml");
        }

        public static void Merge(XSpellShape a, XSpellShape b)
        {
            foreach (var flag in b.Flags) { a.Flags.Add(flag); }
        }

        public static List<XSpellShape> GetPurgedShapes(IEnumerable<XSpellShape> shapes)
        {
            var purgedShapes = new List<XSpellShape>();
            foreach (var shape in shapes)
            {
                var repeated = purgedShapes.FirstOrDefault(s => s.Id == shape.Id);
                if (repeated != null) {
                    repeated.MergeWith(shape);
                }
                else
                {
                    purgedShapes.Add(shape);
                }
            }

            return purgedShapes;
        }

        public static void CleanFile(string path)
        {
            try
            {
                var text = File.ReadAllText(path);
                text = text.Replace("AngleSmall", "WidthSmall");
                text = text.Replace("AngleMedium", "WidthMedium");
                text = text.Replace("AngleLarge", "WidthLarge");
                text = text.Replace("HeightSmall", "ReachSmall");
                text = text.Replace("HeightMedium", "ReachMedium");
                text = text.Replace("HeightLarge", "ReachLarge");
                text = text.Replace("RadiusSmall", "ReachSmall");
                text = text.Replace("RadiusMedium", "ReachMedium");
                text = text.Replace("RadiusLarge", "ReachLarge");
                text = text.Replace("<Flags>TargetedOnOther</Flags>", "");
                File.WriteAllText(path, text);
            }
            catch
            {
                Logging.Write(Common.ErrorColor, "[Zekken] Failed to clean file: " + path);
            }
        }
    }
}