﻿using System;
using System.Collections.Concurrent;
using System.IO;
using System.Windows.Media;
using ff14bot.Helpers;
using ProtoBuf;

namespace Zekken.Avoidance
{
    [Flags]
    public enum SpellShape
    {
        None = 0,

        Cone = 1 << 0,

        Rectangle = 1 << 1,
        WidthSmall = 1 << 2,
        WidthMedium = 1 << 3,
        WidthLarge = 1 << 4,

        Circle = 1 << 5,

        ReachSmall = 1 << 6,
        ReachMedium = 1 << 7,
        ReachLarge = 1 << 8,

        DirForward = 1 << 9,
        DirBackward = 1 << 10,
        DirLeft = 1 << 11,
        DirRight = 1 << 12,

        Interruptible = 1 << 13,

        TargetedOnTarget = 1 << 14,
        TargetedFromCaster = 1 << 15,

        NotTelegraphed = 1 << 16
    }

    public class SpellShapeDatabase
    {
        private Color errorColor = Color.FromRgb(255, 66, 66);
        private Color logColor = Color.FromRgb(201, 115, 255);
        private ConcurrentDictionary<uint, SpellShape> shapes = new ConcurrentDictionary<uint, SpellShape>();

        public bool Load()
        {
            return Load(Common.DEFAULT_DATA_PATH);
        }

        public bool Load(string path)
        {
            if (!File.Exists(path))
            {
                Logging.Write(errorColor, "[Zekken] Spell data file does not exist.");
                return false;
            }

            ConcurrentDictionary<uint, SpellShape> loadedShapes = null;
            using (var stream = new FileStream(path, FileMode.Open, FileAccess.Read))
            {
                try { loadedShapes = Serializer.Deserialize<ConcurrentDictionary<uint, SpellShape>>(stream); }
                catch { Logging.Write(errorColor, "[Zekken] Failed to load spell shape data."); }
            }

            if (loadedShapes == null) { return false; }

            shapes = loadedShapes;
            Logging.Write(logColor, "[Zekken] Loaded spell shape data.");

            return true;
        }

        public bool Save()
        {
            return Save(Common.DEFAULT_DATA_PATH);
        }

        public bool Save(string path)
        {
            var fileInfo = new FileInfo(path);

            if (fileInfo.Directory == null || !Directory.Exists(fileInfo.Directory.FullName))
            {
                Logging.Write(errorColor, "[Zekken] Cannot save spell data, the provided directory does not exist.");
                return false;
            }

            if (fileInfo.Exists)
            {
                try { File.Delete(path); }
                catch
                {
                    Logging.Write(errorColor, "[Zekken] Failed to overwrite spell shape data file.");
                    return false;
                }
            }

            using (var stream = new FileStream(path, FileMode.Create, FileAccess.Write))
            {
                try { Serializer.Serialize(stream, shapes); }
                catch
                {
                    Logging.Write(errorColor, "[Zekken] Failed to save spell shape data.");
                    return false;
                }
            }

            Logging.Write(logColor, "[Zekken] Saved spell shape data to " + path);
            return true;
        }

        public bool TryGetShape(uint spellId, out SpellShape shape)
        {
            return shapes.TryGetValue(spellId, out shape);
        }

        public void SetShape(uint spellId, SpellShape shape)
        {
            shapes[spellId] = shape;
        }

        public bool Contains(uint spellId)
        {
            return shapes.ContainsKey(spellId);
        }
    }
}